#version 330

uniform sampler2D texDiffuseColor;
uniform sampler2D texFragPosition;
uniform sampler2D texFragNormal;
uniform sampler2D texLightData;

uniform vec3 lightPosition;
uniform vec3 ambientColor;
uniform vec3 diffuseColor;
uniform vec3 specularColor;
uniform vec3 attenuation;

uniform vec3 eyePosition;
uniform int lightCount;

smooth in vec2 fragmentUV;

out vec4 fragColor;

#define LIGHT_TYPE_POINT_LIGHT 0.f
#define LIGHT_TYPE_RANGE_LIGHT 1.f
#define LIGHT_TYPE_LASER_LIGHT 2.f

#define MAX_LIGHT_COUNT 500

vec4 shadePointLight(vec3 fragPosition, vec3 fragNormal, vec4 fragDiffuse, vec3 lightPos, vec3 ambientColor, vec3 diffuseColor, vec3 specularColor, vec3 attenuation) {
	vec4 color = vec4(0.0, 0.0, 0.0, 1.0);

	float fragLightDist;
	vec3 fragLightDir;
	float luminocity;
	float diffuseCoefficient;
	float specular;

	fragLightDist = distance(lightPos.xyz, fragPosition);
	fragLightDir = normalize(lightPos.xyz - fragPosition);

	luminocity = 1.0f / (attenuation.x + attenuation.y * fragLightDist
			+ attenuation.z * fragLightDist * fragLightDist);

	// limit (cut off luminocities below that)
	if(luminocity < 0.005f) {
		return color;
	}

	// add ambient color of this light source
	color.rgb += ambientColor;
	
	// dot product of normal and lightdir for phong diffuse light
	diffuseCoefficient = max(0.0, dot(fragNormal, fragLightDir));
	
	// add diffuse light to ambient light
	color.rgb += diffuseCoefficient * diffuseColor;

	color.rgb *= luminocity * fragDiffuse.rgb;
	color.a = fragDiffuse.a;

	vec3 viewDirection = normalize(eyePosition - fragPosition);

	specular = max(0.0, dot(reflect(-fragLightDir, fragNormal), viewDirection));

	if (diffuseCoefficient != 0)
	{
		specular = pow(specular, 128.0) * luminocity;
		color.rgb += specularColor * specular;
	}

	return color;
}

vec4 shadeRangeLight(vec3 fragPosition, vec4 fragDiffuse, vec3 lightPos, float distanceLimit, vec3 ambientColor) {
	vec4 color = vec4(0.0, 0.0, 0.0, 0.0);

	float fragLightDist;
	float distanceFactor;

	fragLightDist = distance(lightPos.xyz, fragPosition);
	distanceFactor = fragLightDist / distanceLimit;

	if(distanceFactor <= 1.f) {
		color = (0.2+distanceFactor*distanceFactor*0.8) * vec4(ambientColor, 1.f) * fragDiffuse;
	}
	return color;
}

vec3 closestPointOnLine(vec3 point, vec3 lineA, vec3 lineB)
{
	vec3 lineDir = lineB - lineA;
	float lineLength = length(lineDir);
	lineDir = normalize(lineDir);

	vec3 v = point - lineA;

	// Project v to LineDirection to get the distance of point from a
	float dist = dot(v, lineDir);
	
	if(dist <= 0) {
		return lineA;
	}
	else {
		if(dist >= lineLength)  {
			return lineB;
		}
		else {
			return lineA + lineDir * dist;
		}
	}
}

void main(void)
{
	vec4 fragDiffuse = texture(texDiffuseColor, fragmentUV);

	fragColor.rgb = vec3(0.f);
	fragColor.a = fragDiffuse.a;

	if(fragColor.a == 0.f) {
		discard;
	}

	vec3 fragPosition = texture(texFragPosition, fragmentUV).xyz;
	vec3 fragNormal = texture(texFragNormal, fragmentUV).xyz;
	
	float stepSizeX = 1.f/7.f;
	float stepSizeY = 1.f/float(MAX_LIGHT_COUNT);

	for(int i = 0; i < lightCount; i++) {
		float v = (float(i)+0.5f)*stepSizeY;
		
		vec3 lightConfig = texture(texLightData, vec2(0.5f*stepSizeX, v)).xyz;

		// on / off
		if(lightConfig.x == 1.f) {

			vec3 posA = texture(texLightData, vec2(1.5f*stepSizeX, v)).xyz;

			float dist = distance(fragPosition, posA);

			vec3 ambientColor = texture(texLightData, vec2(3.5f*stepSizeX, v)).xyz;
			
			if(lightConfig.y == LIGHT_TYPE_POINT_LIGHT) {
				vec3 diffuseColor = texture(texLightData, vec2(4.5f*stepSizeX, v)).xyz;
				vec3 specularColor = texture(texLightData, vec2(5.5f*stepSizeX, v)).xyz;
				vec3 attenuations = texture(texLightData, vec2(6.5f*stepSizeX, v)).xyz;
				
				fragColor += shadePointLight(fragPosition, fragNormal, fragDiffuse, posA, ambientColor, diffuseColor, specularColor, attenuations);
			}
			else if(lightConfig.y == LIGHT_TYPE_RANGE_LIGHT) {
				float distanceLimit = lightConfig.z;

				fragColor += shadeRangeLight(fragPosition, fragDiffuse, posA, distanceLimit, ambientColor);
			}
			else if(lightConfig.y == LIGHT_TYPE_LASER_LIGHT) {
				vec3 posB = texture(texLightData, vec2(2.5f*stepSizeX, v)).xyz;
				vec3 diffuseColor = texture(texLightData, vec2(4.5f*stepSizeX, v)).xyz;
				vec3 specularColor = texture(texLightData, vec2(5.5f*stepSizeX, v)).xyz;
				vec3 attenuations = texture(texLightData, vec2(6.5f*stepSizeX, v)).xyz;
				
				fragColor += shadePointLight(fragPosition, fragNormal, fragDiffuse, closestPointOnLine(fragPosition, posA, posB), ambientColor, diffuseColor, specularColor, attenuations);
			}
		}
	}
}
